# clean

"""Postprocessing option computes overall consumption, production,self-consumption and injection as well as selfconsumption rate and self-sufficiency rate.

KPis for PV-battery systems in houses:
https://solar.htw-berlin.de/wp-content/uploads/WENIGER-2017-Vergleich-verschiedener-Kennzahlen-zur-Bewertung-von-PV-Batteriesystemen.pdf.
"""

import os
from typing import List, Tuple, Dict, Optional
from pathlib import Path
import pandas as pd
from hisim.component import ComponentOutput
from hisim.component_wrapper import ComponentWrapper
from hisim.loadtypes import ComponentType, InandOutputType, DistrictNames
from hisim import log
from hisim.components.electricity_meter import ElectricityMeter
from hisim.postprocessing.postprocessing_datatransfer import PostProcessingDataTransfer
from hisim.postprocessing.kpi_computation.kpi_structure import KpiTagEnumClass, KpiEntry


class KpiPreparation:
    """Class for generating and calculating key performance indicators."""

    def __init__(
        self, post_processing_data_transfer: PostProcessingDataTransfer, building_objects_in_district_list: list
    ):
        """Initialize further variables."""
        self.post_processing_data_transfer = post_processing_data_transfer
        self.building_objects_in_district_list = building_objects_in_district_list
        self.kpi_collection_dict_unsorted: Dict = {}
        # get important variables
        self.wrapped_components = self.post_processing_data_transfer.wrapped_components
        self.results = self.post_processing_data_transfer.results
        self.all_outputs = self.post_processing_data_transfer.all_outputs
        self.simulation_parameters = self.post_processing_data_transfer.simulation_parameters
        self.get_all_component_kpis(wrapped_components=self.wrapped_components)

    def filter_results_according_to_postprocessing_flags(
        self,
        all_outputs: List,
        results: pd.DataFrame,
        building_objects_in_district: str,
    ) -> pd.DataFrame:
        """Filter results according to postprocessing flags and get consumption, production, battery charge and battery discharge.

        Also evaluates battery charge and discharge, because it is relevant for self consumption rates.
        To be recognised as production the connected outputs need a postprocessing flag: InandOutputType.ELECTRICITY_PRODUCTION,
        consumption is flagged with either InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED or InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
        storage charge/discharge is flagged with InandOutputType.CHARGE_DISCHARGE. For batteries to be considered as wished, they additionally need the
        Component itself as postprocesessing flag: ComponentType.CAR_BATTERY or ComponentType.BATTERY
        """

        # initialize columns consumption, production, battery_charge, battery_discharge, storage
        total_consumption_ids = []
        total_production_ids = []
        buildings_production_ids = []
        pv_production_ids = []
        windturbine_production_ids = []
        battery_charge_discharge_ids = []
        buildings_consumption_ids = []

        index: int
        output: ComponentOutput

        for index, output in enumerate(all_outputs):
            if (building_objects_in_district == output.component_name.split("_")[0] or
                    not self.simulation_parameters.multiple_buildings):
                if output.postprocessing_flag is not None:
                    if InandOutputType.ELECTRICITY_PRODUCTION in output.postprocessing_flag:
                        total_production_ids.append(index)
                        if ComponentType.PV in output.postprocessing_flag:
                            pv_production_ids.append(index)
                        elif ComponentType.WINDTURBINE in output.postprocessing_flag:
                            windturbine_production_ids.append(index)
                        elif ComponentType.BUILDINGS in output.postprocessing_flag:
                            buildings_production_ids.append(index)
                    if (
                        InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED in output.postprocessing_flag
                        or InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED in output.postprocessing_flag
                    ):
                        total_consumption_ids.append(index)
                        log.debug(
                            "Output considered in total electricity consumption "
                            + output.full_name
                            + " "
                            + str(output.unit)
                        )
                        if ComponentType.BUILDINGS in output.postprocessing_flag:
                            buildings_consumption_ids.append(index)

                    if InandOutputType.CHARGE_DISCHARGE in output.postprocessing_flag:
                        if ComponentType.BATTERY in output.postprocessing_flag:
                            battery_charge_discharge_ids.append(index)
                        elif ComponentType.CAR_BATTERY in output.postprocessing_flag:
                            total_consumption_ids.append(index)
                            log.debug(
                                "Output considered in total electricity consumption "
                                + output.full_name
                                + " "
                                + str(output.unit)
                            )
                else:
                    continue

        result_dataframe = pd.DataFrame()
        result_dataframe["total_consumption"] = (
            pd.DataFrame(results.iloc[:, total_consumption_ids]).clip(lower=0).sum(axis=1)
        )
        result_dataframe["building_consumption"] = (
            pd.DataFrame(results.iloc[:, buildings_consumption_ids]).clip(lower=0).sum(axis=1)
        )
        result_dataframe["total_production"] = (
            pd.DataFrame(results.iloc[:, total_production_ids]).clip(lower=0).sum(axis=1)
        )
        result_dataframe["building_production"] = (
            pd.DataFrame(results.iloc[:, buildings_production_ids]).clip(lower=0).sum(axis=1)
        )
        result_dataframe["pv_production"] = pd.DataFrame(results.iloc[:, pv_production_ids]).clip(lower=0).sum(axis=1)
        result_dataframe["windturbine_production"] = (
            pd.DataFrame(results.iloc[:, windturbine_production_ids]).clip(lower=0).sum(axis=1)
        )
        result_dataframe["battery_charge"] = (
            pd.DataFrame(results.iloc[:, battery_charge_discharge_ids]).clip(lower=0).sum(axis=1)
        )
        result_dataframe["battery_discharge"] = pd.DataFrame(results.iloc[:, battery_charge_discharge_ids]).clip(
            upper=0
        ).sum(axis=1) * (-1)

        return result_dataframe

    def compute_total_energy_from_power_timeseries(
        self, power_timeseries_in_watt: pd.Series, timeresolution: int
    ) -> float:
        """Computes the energy in kWh from a power timeseries in W."""
        if power_timeseries_in_watt.empty:
            return 0.0

        energy_in_kilowatt_hour = float(power_timeseries_in_watt.sum() * timeresolution / 3.6e6)
        return energy_in_kilowatt_hour

    def compute_electricity_consumption_and_production_and_battery_kpis(
        self,
        result_dataframe: pd.DataFrame,
        building_objects_in_district: str,
        kpi_tag: KpiTagEnumClass,
    ) -> Tuple[float, float, float, float, float]:
        """Compute electricity consumption and production and battery kpis."""

        # sum consumption and production over time
        total_electricity_consumption_in_kilowatt_hour = round(
            self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["total_consumption"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            ),
            1,
        )
        building_electricity_consumption_in_kilowatt_hour = round(
            self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["building_consumption"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            ),
            1,
        )

        total_electricity_production_in_kilowatt_hour = round(
            self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["total_production"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            ),
            1,
        )
        pv_production_in_kilowatt_hour = round(
            self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["pv_production"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            ),
            1,
        )
        windturbine_production_in_kilowatt_hour = round(
            self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["windturbine_production"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            ),
            1,
        )
        building_production_in_kilowatt_hour = round(
            self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["building_production"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            ),
            1,
        )

        # compute battery kpis
        (
            battery_charging_energy_in_kilowatt_hour,
            battery_discharging_energy_in_kilowatt_hour,
            battery_losses_in_kilowatt_hour,
        ) = self.compute_battery_kpis(result_dataframe=result_dataframe)

        # if battery losses are not zero, add to total consumption because this is what is consumed by battery indepently from charging and discharging
        total_electricity_consumption_in_kilowatt_hour = (
            total_electricity_consumption_in_kilowatt_hour + battery_losses_in_kilowatt_hour
        )
        log.debug("Battery losses " + str(battery_losses_in_kilowatt_hour))
        log.debug(
            "Total electricity consumption (battery losses included) "
            + str(total_electricity_consumption_in_kilowatt_hour)
        )

        # make kpi entry
        total_consumtion_entry = KpiEntry(
            name="Total electricity consumption",
            unit="kWh",
            value=total_electricity_consumption_in_kilowatt_hour,
            tag=kpi_tag,
        )

        total_production_entry = KpiEntry(
            name="Total electricity production",
            unit="kWh",
            value=total_electricity_production_in_kilowatt_hour,
            tag=kpi_tag,
        )
        pv_production_entry = KpiEntry(
            name="PV production", unit="kWh", value=pv_production_in_kilowatt_hour, tag=kpi_tag
        )
        windturbine_production_entry = KpiEntry(
            name="Windturbine production", unit="kWh", value=windturbine_production_in_kilowatt_hour, tag=kpi_tag
        )
        if any(word in building_objects_in_district for word in DistrictNames):
            building_consumption_entry = KpiEntry(
                name="Total building electricity consumption",
                unit="kWh",
                value=building_electricity_consumption_in_kilowatt_hour,
                tag=kpi_tag,
            )

            building_production_entry = KpiEntry(
                name="Building production", unit="kWh", value=building_production_in_kilowatt_hour, tag=kpi_tag
            )
        battery_charging_entry = KpiEntry(
            name="Battery charging energy",
            unit="kWh",
            value=battery_charging_energy_in_kilowatt_hour,
            tag=KpiTagEnumClass.BATTERY,
        )
        battery_discharging_entry = KpiEntry(
            name="Battery discharging energy",
            unit="kWh",
            value=battery_discharging_energy_in_kilowatt_hour,
            tag=KpiTagEnumClass.BATTERY,
        )
        battery_losses_entry = KpiEntry(
            name="Battery losses",
            unit="kWh",
            value=battery_losses_in_kilowatt_hour,
            tag=KpiTagEnumClass.BATTERY,
        )

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[building_objects_in_district].update(
            {
                total_consumtion_entry.name: total_consumtion_entry.to_dict(),
                total_production_entry.name: total_production_entry.to_dict(),
                pv_production_entry.name: pv_production_entry.to_dict(),
                windturbine_production_entry.name: windturbine_production_entry.to_dict(),
                battery_charging_entry.name: battery_charging_entry.to_dict(),
                battery_discharging_entry.name: battery_discharging_entry.to_dict(),
                battery_losses_entry.name: battery_losses_entry.to_dict(),
            }
        )
        if any(word in building_objects_in_district for word in DistrictNames):
            self.kpi_collection_dict_unsorted[building_objects_in_district].update(
                {
                    building_production_entry.name: building_production_entry.to_dict(),
                    building_consumption_entry.name: building_consumption_entry.to_dict(),
                }
            )

        return (
            total_electricity_consumption_in_kilowatt_hour,
            total_electricity_production_in_kilowatt_hour,
            pv_production_in_kilowatt_hour,
            windturbine_production_in_kilowatt_hour,
            building_production_in_kilowatt_hour,
        )

    def compute_self_consumption_injection_self_sufficiency(
        self,
        result_dataframe: pd.DataFrame,
        electricity_production_in_kilowatt_hour: float,
        electricity_consumption_in_kilowatt_hour: float,
        building_objects_in_district: str,
        kpi_tag: KpiTagEnumClass,
    ) -> pd.DataFrame:
        """Computes the self consumption, grid injection, self-sufficiency and battery losses if electricty production is bigger than zero."""

        if electricity_production_in_kilowatt_hour > 0:
            # account for battery
            production_with_battery = result_dataframe["total_production"] + result_dataframe["battery_discharge"]
            consumption_with_battery = result_dataframe["total_consumption"] + result_dataframe["battery_charge"]

            # evaluate injection and sum over time
            grid_injection_series_in_watt: pd.Series = production_with_battery - consumption_with_battery

            # evaluate self consumption and immidiately sum over time
            # battery is charged (counting to consumption) and discharged (counting to production)
            # -> only one direction can be counted, otherwise the self-consumption can be greater than 100.
            # Here the production side is counted (battery_discharge).
            self_consumption_series_in_watt: pd.Series = (
                pd.concat(
                    (
                        production_with_battery[production_with_battery <= result_dataframe["total_consumption"]],
                        result_dataframe["total_consumption"][
                            result_dataframe["total_consumption"] < production_with_battery
                        ],
                    )
                )
                .groupby(level=0)
                .sum()
            )

            grid_injection_in_kilowatt_hour = self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=grid_injection_series_in_watt[grid_injection_series_in_watt > 0],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            )

            self_consumption_in_kilowatt_hour = self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=self_consumption_series_in_watt,
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            )

            # compute self consumption rate and autarkie rate
            self_consumption_rate_in_percent = 100 * (
                self_consumption_in_kilowatt_hour / electricity_production_in_kilowatt_hour
            )
            self_sufficiency_rate_in_percent = 100 * (
                self_consumption_in_kilowatt_hour / electricity_consumption_in_kilowatt_hour
            )
            if self_sufficiency_rate_in_percent > 100:
                raise ValueError(
                    "The self-sufficiency rate should not be over 100 %. Something is wrong here. Please check your code."
                )

        else:
            self_consumption_series_in_watt = pd.Series([])
            grid_injection_series_in_watt = pd.Series([])
            self_consumption_in_kilowatt_hour = 0
            grid_injection_in_kilowatt_hour = 0
            self_consumption_rate_in_percent = 0
            self_sufficiency_rate_in_percent = 0

        # add injection and self-consumption timeseries to result dataframe
        result_dataframe["self_consumption_in_watt"] = self_consumption_series_in_watt
        result_dataframe["grid_injection_in_watt"] = grid_injection_series_in_watt

        # make kpi entry
        grid_injection_entry = KpiEntry(
            name="Grid injection of electricity", unit="kWh", value=grid_injection_in_kilowatt_hour, tag=kpi_tag
        )
        self_consumption_entry = KpiEntry(
            name="Self-consumption of electricity", unit="kWh", value=self_consumption_in_kilowatt_hour, tag=kpi_tag
        )
        self_consumption_rate_entry = KpiEntry(
            name="Self-consumption rate of electricity", unit="%", value=self_consumption_rate_in_percent, tag=kpi_tag
        )
        self_sufficiency_rate_entry = KpiEntry(
            name="Self-sufficiency rate of electricity", unit="%", value=self_sufficiency_rate_in_percent, tag=kpi_tag
        )

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[building_objects_in_district].update(
            {
                grid_injection_entry.name: grid_injection_entry.to_dict(),
                self_consumption_entry.name: self_consumption_entry.to_dict(),
                self_consumption_rate_entry.name: self_consumption_rate_entry.to_dict(),
                self_sufficiency_rate_entry.name: self_sufficiency_rate_entry.to_dict(),
            }
        )
        return result_dataframe

    def compute_battery_kpis(self, result_dataframe: pd.DataFrame) -> Tuple[float, float, float]:
        """Compute battery kpis."""

        if not result_dataframe["battery_charge"].empty:
            battery_charging_energy_in_kilowatt_hour = self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["battery_charge"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            )
            battery_discharging_energy_in_kilowatt_hour = self.compute_total_energy_from_power_timeseries(
                power_timeseries_in_watt=result_dataframe["battery_discharge"],
                timeresolution=self.simulation_parameters.seconds_per_timestep,
            )
            battery_losses_in_kilowatt_hour = (
                battery_charging_energy_in_kilowatt_hour - battery_discharging_energy_in_kilowatt_hour
            )
        else:
            battery_charging_energy_in_kilowatt_hour = 0.0
            battery_discharging_energy_in_kilowatt_hour = 0.0
            battery_losses_in_kilowatt_hour = 0.0

        return (
            battery_charging_energy_in_kilowatt_hour,
            battery_discharging_energy_in_kilowatt_hour,
            battery_losses_in_kilowatt_hour,
        )

    def get_electricity_to_and_from_grid_from_electricty_meter(
        self,
        building_objects_in_district: str,
    ) -> Tuple[Optional[float], Optional[float]]:
        """Get electricity to and from grid from electricity meter."""

        total_energy_from_grid_in_kwh: Optional[float] = None
        total_energy_to_grid_in_kwh: Optional[float] = None
        for kpi_entry in self.kpi_collection_dict_unsorted[building_objects_in_district].values():
            if (
                isinstance(kpi_entry["description"], str)
                and ElectricityMeter.get_classname() in kpi_entry["description"]
            ):
                if kpi_entry["name"] == "Total energy from grid" and kpi_entry["unit"] == "kWh":
                    total_energy_from_grid_in_kwh = kpi_entry["value"]
                elif kpi_entry["name"] == "Total energy to grid" and kpi_entry["unit"] == "kWh":
                    total_energy_to_grid_in_kwh = kpi_entry["value"]
            else:
                continue

        return total_energy_from_grid_in_kwh, total_energy_to_grid_in_kwh

    def compute_relative_electricity_demand(
        self,
        total_electricity_consumption_in_kilowatt_hour: float,
        electricity_from_grid_in_kilowatt_hour: Optional[float],
        building_objects_in_district: str,
        kpi_tag: KpiTagEnumClass,
        name: str = "Relative electricity demand from grid",
    ) -> Optional[float]:
        """Return the relative electricity demand."""

        if electricity_from_grid_in_kilowatt_hour is None:
            relative_electricity_demand_from_grid_in_percent = None
        else:
            relative_electricity_demand_from_grid_in_percent = (
                round(electricity_from_grid_in_kilowatt_hour, 1)
                / round(total_electricity_consumption_in_kilowatt_hour, 1)
                * 100
            )
            if relative_electricity_demand_from_grid_in_percent > 100:
                raise ValueError(
                    "The relative elecricity demand should not be over 100 %. Something is wrong here. Please check your code."
                    f"Electricity from grid {electricity_from_grid_in_kilowatt_hour} kWh, "
                    f"total electricity consumption {total_electricity_consumption_in_kilowatt_hour} kWh."
                )

        # make kpi entry
        relative_electricity_demand_entry = KpiEntry(
            name=name, unit="%", value=relative_electricity_demand_from_grid_in_percent, tag=kpi_tag
        )

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[building_objects_in_district].update(
            {relative_electricity_demand_entry.name: relative_electricity_demand_entry.to_dict()}
        )

        return relative_electricity_demand_from_grid_in_percent

    def compute_self_sufficiency_according_to_solar_htw_berlin(
        self,
        relative_electricty_demand_in_percent: Optional[float],
        building_objects_in_district: str,
        kpi_tag: KpiTagEnumClass,
        name: str = "Self-sufficiency rate according to solar htw berlin",
    ) -> None:
        """Return the self-sufficiency rate according to solar htw berlin.

        https://solar.htw-berlin.de/wp-content/uploads/WENIGER-2017-Vergleich-verschiedener-Kennzahlen-zur-Bewertung-von-PV-Batteriesystemen.pdf.
        """
        if relative_electricty_demand_in_percent is None:
            self_sufficiency_rate_in_percent = None
        else:
            self_sufficiency_rate_in_percent = 100 - relative_electricty_demand_in_percent
            if self_sufficiency_rate_in_percent > 100:
                raise ValueError(
                    "The self-sufficiency rate should not be over 100 %. Something is wrong here. Please check your code. "
                    f"The realtive electricity demand is {relative_electricty_demand_in_percent} %. "
                )

        # make kpi entry
        self_sufficiency_rate_entry = KpiEntry(name=name, unit="%", value=self_sufficiency_rate_in_percent, tag=kpi_tag)

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[building_objects_in_district].update(
            {self_sufficiency_rate_entry.name: self_sufficiency_rate_entry.to_dict()}
        )

    def compute_self_consumption_rate_according_to_solar_htw_berlin(
        self,
        total_electricity_production_in_kilowatt_hour: float,
        electricity_to_grid_in_kilowatt_hour: Optional[float],
        building_objects_in_district: str,
        kpi_tag: KpiTagEnumClass,
        name: str = "Self-consumption rate according to solar htw berlin",
    ) -> None:
        """Return self-consumption according to solar htw berlin.

        https://solar.htw-berlin.de/wp-content/uploads/WENIGER-2017-Vergleich-verschiedener-Kennzahlen-zur-Bewertung-von-PV-Batteriesystemen.pdf.
        https://academy.dualsun.com/hc/en-us/articles/360018456939-How-is-the-self-consumption-rate-calculated-on-MyDualSun.
        """
        if electricity_to_grid_in_kilowatt_hour is None or total_electricity_production_in_kilowatt_hour == 0:
            self_consumption_rate_in_percent = None
        else:
            self_consumption_rate_in_percent = (
                (total_electricity_production_in_kilowatt_hour - electricity_to_grid_in_kilowatt_hour)
                / total_electricity_production_in_kilowatt_hour
                * 100
            )
            if self_consumption_rate_in_percent > 100:
                raise ValueError(
                    "The self-consumption rate should not be over 100 %. Something is wrong here. Please check your code."
                    f"Electricity to grid {electricity_to_grid_in_kilowatt_hour} kWh, "
                    f"total electricity production {total_electricity_production_in_kilowatt_hour} kWh."
                )

        # make kpi entry
        self_consumption_rate_entry = KpiEntry(
            name=name,
            unit="%",
            value=self_consumption_rate_in_percent,
            tag=kpi_tag,
        )

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[building_objects_in_district].update(
            {self_consumption_rate_entry.name: self_consumption_rate_entry.to_dict()}
        )

    def compute_ratio_between_two_values_and_set_as_kpi(
        self,
        denominator_value: float,
        numerator_value: float,
        kpi_name: str,
        building_objects_in_district: str,
        kpi_tag: KpiTagEnumClass,
    ) -> None:
        """Compute the ratio of two values.

        ratio = denominator / numerator * 100 [%].
        """
        ratio_in_percent = denominator_value / numerator_value * 100
        # make kpi entry
        ratio_in_percent_entry = KpiEntry(
            name=kpi_name,
            unit="%",
            value=ratio_in_percent,
            tag=kpi_tag,
        )

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[building_objects_in_district].update(
            {ratio_in_percent_entry.name: ratio_in_percent_entry.to_dict()}
        )

    def read_opex_and_capex_costs_from_results(self, building_object: str) -> None:
        """Get CAPEX and OPEX costs for simulated period.

        This function will read the opex and capex costs from the results.
        """
        # get costs and emissions from electricity meter and gas meter
        electricity_costs_in_euro: float = 0
        electricity_co2_in_kg: float = 0
        gas_costs_in_euro: float = 0
        gas_co2_in_kg: float = 0
        heating_costs_in_euro: float = 0
        heating_co2_in_kg: float = 0

        for kpi_name, kpi_entry in self.kpi_collection_dict_unsorted[building_object].items():
            if kpi_entry["tag"] == KpiTagEnumClass.ELECTRICITY_METER.value:
                if kpi_name == "Opex costs of electricity consumption from grid":
                    electricity_costs_in_euro = kpi_entry["value"]
                if kpi_name == "CO2 footprint of electricity consumption from grid":
                    electricity_co2_in_kg = kpi_entry["value"]

            elif kpi_entry["tag"] == KpiTagEnumClass.GAS_METER.value:
                if kpi_name == "Opex costs of gas consumption from grid":
                    gas_costs_in_euro = kpi_entry["value"]
                if kpi_name == "CO2 footprint of gas consumption from grid":
                    gas_co2_in_kg = kpi_entry["value"]

            elif kpi_entry["tag"] == KpiTagEnumClass.HEATING_METER.value:
                if kpi_name == "Opex costs of heat consumption from grid":
                    heating_costs_in_euro = kpi_entry["value"]
                if kpi_name == "CO2 footprint of heat consumption from grid":
                    heating_co2_in_kg = kpi_entry["value"]

        # get CAPEX and OPEX costs for simulated period
        capex_results_path = os.path.join(
            self.simulation_parameters.result_directory, "investment_cost_co2_footprint.csv"
        )
        opex_results_path = os.path.join(
            self.simulation_parameters.result_directory, "operational_costs_co2_footprint.csv"
        )
        if Path(opex_results_path).exists():
            opex_df = pd.read_csv(opex_results_path, index_col=0, sep=";")
            log.debug("Opex df " + str(opex_df) + "\n")
            if self.simulation_parameters.multiple_buildings:
                total_maintenance_cost_per_simulated_period = opex_df["Maintenance Costs in EUR"].loc[
                    building_object + "_Total"
                ]
                total_maintenance_cost_per_simulated_period_without_hp = opex_df["Maintenance Costs in EUR"].loc[
                    building_object + "_Total_without_heatpump"
                ]
                total_maintenance_cost_per_simulated_period_only_hp = opex_df["Maintenance Costs in EUR"].loc[
                    building_object + "_Total_only_heatpump"
                ]
            if not self.simulation_parameters.multiple_buildings:
                total_maintenance_cost_per_simulated_period = opex_df["Maintenance Costs in EUR"].loc[
                    "Total"
                ]
                total_maintenance_cost_per_simulated_period_without_hp = opex_df["Maintenance Costs in EUR"].loc[
                    "Total_without_heatpump"
                ]
                total_maintenance_cost_per_simulated_period_only_hp = opex_df["Maintenance Costs in EUR"].loc[
                    "Total_only_heatpump"
                ]

        else:
            log.warning("OPEX-costs for components are not calculated yet. Set PostProcessingOptions.COMPUTE_OPEX")
            total_maintenance_cost_per_simulated_period = 0
            total_maintenance_cost_per_simulated_period_without_hp = 0
            total_maintenance_cost_per_simulated_period_only_hp = 0

        if Path(capex_results_path).exists():
            capex_df = pd.read_csv(capex_results_path, index_col=0, sep=";")
            log.debug("Capex df " + str(capex_df) + "\n")
            if self.simulation_parameters.multiple_buildings:
                total_investment_cost_per_simulated_period = capex_df["Investment in EUR"].loc[
                    building_object + "_Total_per_simulated_period"
                ]
                total_device_co2_footprint_per_simulated_period = capex_df["Device CO2-footprint in kg"].loc[
                    building_object + "_Total_per_simulated_period"
                ]
                total_investment_cost_per_simulated_period_without_hp = capex_df["Investment in EUR"].loc[
                    building_object + "_Total_per_simulated_period_without_heatpump"
                ]
                total_device_co2_footprint_per_simulated_period_without_hp = capex_df["Device CO2-footprint in kg"].loc[
                    building_object + "_Total_per_simulated_period_without_heatpump"
                ]
                total_investment_cost_per_simulated_period_only_hp = capex_df["Investment in EUR"].loc[
                    building_object + "_Total_per_simulated_period_only_heatpump"
                ]
                total_device_co2_footprint_per_simulated_period_only_hp = capex_df["Device CO2-footprint in kg"].loc[
                    building_object + "_Total_per_simulated_period_only_heatpump"
                ]
            if not self.simulation_parameters.multiple_buildings:
                total_investment_cost_per_simulated_period = capex_df["Investment in EUR"].loc[
                    "Total_per_simulated_period"
                ]
                total_device_co2_footprint_per_simulated_period = capex_df["Device CO2-footprint in kg"].loc[
                    "Total_per_simulated_period"
                ]
                total_investment_cost_per_simulated_period_without_hp = capex_df["Investment in EUR"].loc[
                    "Total_per_simulated_period_without_heatpump"
                ]
                total_device_co2_footprint_per_simulated_period_without_hp = capex_df["Device CO2-footprint in kg"].loc[
                    "Total_per_simulated_period_without_heatpump"
                ]
                total_investment_cost_per_simulated_period_only_hp = capex_df["Investment in EUR"].loc[
                    "Total_per_simulated_period_only_heatpump"
                ]
                total_device_co2_footprint_per_simulated_period_only_hp = capex_df["Device CO2-footprint in kg"].loc[
                    "Total_per_simulated_period_only_heatpump"
                ]
        else:
            log.warning("CAPEX-costs for components are not calculated yet. Set PostProcessingOptions.COMPUTE_CAPEX")
            total_investment_cost_per_simulated_period = 0
            total_device_co2_footprint_per_simulated_period = 0
            total_investment_cost_per_simulated_period_without_hp = 0
            total_device_co2_footprint_per_simulated_period_without_hp = 0
            total_investment_cost_per_simulated_period_only_hp = 0
            total_device_co2_footprint_per_simulated_period_only_hp = 0

        # make kpi entry
        total_electricity_costs_entry = KpiEntry(
            name="Costs of grid electricity for simulated period",
            unit="EUR",
            value=electricity_costs_in_euro,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_electricity_co2_footprint_entry = KpiEntry(
            name="CO2 footprint of grid electricity for simulated period",
            unit="kg",
            value=electricity_co2_in_kg,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )
        total_gas_costs_entry = KpiEntry(
            name="Costs of grid gas for simulated period",
            unit="EUR",
            value=gas_costs_in_euro,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_gas_co2_emissions_entry = KpiEntry(
            name="CO2 footprint of grid gas for simulated period",
            unit="kg",
            value=gas_co2_in_kg,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )
        total_heat_costs_entry = KpiEntry(
            name="Costs of grid heat for simulated period",
            unit="EUR",
            value=heating_costs_in_euro,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_heat_co2_emissions_entry = KpiEntry(
            name="CO2 footprint of grid heat consumption for simulated period",
            unit="kg",
            value=heating_co2_in_kg,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )
        total_investment_cost_per_simulated_period_entry = KpiEntry(
            name="Investment costs for equipment per simulated period",
            unit="EUR",
            value=total_investment_cost_per_simulated_period,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_device_co2_footprint_per_simulated_period_entry = KpiEntry(
            name="CO2 footprint for equipment per simulated period",
            unit="kg",
            value=total_device_co2_footprint_per_simulated_period,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )
        total_maintenance_cost_entry = KpiEntry(
            name="Maintenance costs for simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )

        total_energy_cost_entry = KpiEntry(
            name="Energy grid costs for simulated period",
            unit="EUR",
            value=gas_costs_in_euro
            + electricity_costs_in_euro
            + heating_costs_in_euro,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )

        total_cost_entry = KpiEntry(
            name="Total costs for simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period
            + total_investment_cost_per_simulated_period
            + gas_costs_in_euro
            + electricity_costs_in_euro
            + heating_costs_in_euro,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_emissions_entry = KpiEntry(
            name="Total CO2 emissions for simulated period",
            unit="kg",
            value=total_device_co2_footprint_per_simulated_period
            + gas_co2_in_kg
            + electricity_co2_in_kg
            + heating_co2_in_kg,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )

        total_investment_cost_per_simulated_period_without_hp_entry = KpiEntry(
            name="Investment costs for equipment without heatpump per simulated period",
            unit="EUR",
            value=total_investment_cost_per_simulated_period_without_hp,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_device_co2_footprint_per_simulated_period_without_hp_entry = KpiEntry(
            name="CO2 footprint for equipment without heatpump per simulated period",
            unit="kg",
            value=total_device_co2_footprint_per_simulated_period_without_hp,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )
        total_maintenance_cost_without_hp_entry = KpiEntry(
            name="Maintenance costs without heatpump for simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_without_hp,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_cost_without_hp_entry = KpiEntry(
            name="Total costs without heatpump for simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_without_hp
            + total_investment_cost_per_simulated_period_without_hp
            + gas_costs_in_euro
            + electricity_costs_in_euro
            + heating_costs_in_euro,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_emissions_without_hp_entry = KpiEntry(
            name="Total CO2 emissions without heatpump for simulated period",
            unit="kg",
            value=total_device_co2_footprint_per_simulated_period_without_hp
            + gas_co2_in_kg
            + electricity_co2_in_kg
            + heating_co2_in_kg,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )

        total_investment_cost_per_simulated_period_only_hp_entry = KpiEntry(
            name="Investment costs for equipment only heatpump per simulated period",
            unit="EUR",
            value=total_investment_cost_per_simulated_period_only_hp,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_device_co2_footprint_per_simulated_period_only_hp_entry = KpiEntry(
            name="CO2 footprint for equipment only heatpump per simulated period",
            unit="kg",
            value=total_device_co2_footprint_per_simulated_period_only_hp,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )
        total_maintenance_cost_only_hp_entry = KpiEntry(
            name="Maintenance costs only heatpump for simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_only_hp,
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_cost_only_hp_entry = KpiEntry(
            name="Total costs only heatpump for simulated period",
            unit="EUR",
            value=(
                total_maintenance_cost_per_simulated_period_only_hp + total_investment_cost_per_simulated_period_only_hp
            ),
            tag=(
                KpiTagEnumClass.COSTS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.COSTS_DISTRICT_GRID
            ),
        )
        total_emissions_only_hp_entry = KpiEntry(
            name="Total CO2 emissions only heatpump for simulated period",
            unit="kg",
            value=total_device_co2_footprint_per_simulated_period_only_hp,
            tag=(
                KpiTagEnumClass.EMISSIONS
                if not any(word in building_object for word in DistrictNames)
                else KpiTagEnumClass.EMISSIONS_DISTRICT_GRID
            ),
        )

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[building_object].update(
            {
                total_electricity_costs_entry.name: total_electricity_costs_entry.to_dict(),
                total_electricity_co2_footprint_entry.name: total_electricity_co2_footprint_entry.to_dict(),
                total_gas_costs_entry.name: total_gas_costs_entry.to_dict(),
                total_gas_co2_emissions_entry.name: total_gas_co2_emissions_entry.to_dict(),
                total_heat_costs_entry.name: total_heat_costs_entry.to_dict(),
                total_heat_co2_emissions_entry.name: total_heat_co2_emissions_entry.to_dict(),
                total_investment_cost_per_simulated_period_entry.name: total_investment_cost_per_simulated_period_entry.to_dict(),
                total_device_co2_footprint_per_simulated_period_entry.name: total_device_co2_footprint_per_simulated_period_entry.to_dict(),
                total_energy_cost_entry.name: total_energy_cost_entry.to_dict(),
                total_maintenance_cost_entry.name: total_maintenance_cost_entry.to_dict(),
                total_cost_entry.name: total_cost_entry.to_dict(),
                total_emissions_entry.name: total_emissions_entry.to_dict(),
                total_investment_cost_per_simulated_period_without_hp_entry.name: total_investment_cost_per_simulated_period_without_hp_entry.to_dict(),
                total_device_co2_footprint_per_simulated_period_without_hp_entry.name: total_device_co2_footprint_per_simulated_period_without_hp_entry.to_dict(),
                total_maintenance_cost_without_hp_entry.name: total_maintenance_cost_without_hp_entry.to_dict(),
                total_cost_without_hp_entry.name: total_cost_without_hp_entry.to_dict(),
                total_emissions_without_hp_entry.name: total_emissions_without_hp_entry.to_dict(),
                total_investment_cost_per_simulated_period_only_hp_entry.name: total_investment_cost_per_simulated_period_only_hp_entry.to_dict(),
                total_device_co2_footprint_per_simulated_period_only_hp_entry.name: total_device_co2_footprint_per_simulated_period_only_hp_entry.to_dict(),
                total_maintenance_cost_only_hp_entry.name: total_maintenance_cost_only_hp_entry.to_dict(),
                total_cost_only_hp_entry.name: total_cost_only_hp_entry.to_dict(),
                total_emissions_only_hp_entry.name: total_emissions_only_hp_entry.to_dict(),
            }
        )

    def create_overall_district_kpi_collection(self, district_name: str) -> Tuple[
        float,
        float,
        float,
    ]:
        """Overall kpis for districts."""
        electricity_consumption_all_single_buildings_in_kilowatt_hour = 0.0
        electricity_produktion_all_single_buildings_in_kilowatt_hour = 0.0
        self_consumption_all_single_buildings_in_kilowatt_hour = 0.0
        electricity_production_district_in_kilowatt_hour = 0.0
        electricity_consumption_district_in_kilowatt_hour = 0.0
        self_consumption_district_in_kilowatt_hour = 0.0

        for building_objects_in_district in self.building_objects_in_district_list:
            if not any(word in building_objects_in_district for word in DistrictNames):
                electricity_consumption_all_single_buildings_in_kilowatt_hour += self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Total electricity consumption"]["value"]
                electricity_produktion_all_single_buildings_in_kilowatt_hour += self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Total electricity production"]["value"]
                self_consumption_all_single_buildings_in_kilowatt_hour += self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Self-consumption of electricity"]["value"]

            if any(word in building_objects_in_district for word in DistrictNames):
                electricity_production_district_in_kilowatt_hour += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district]["Total electricity production"][
                        "value"
                    ]
                    - self.kpi_collection_dict_unsorted[building_objects_in_district]["Building production"]["value"]
                )
                electricity_consumption_district_in_kilowatt_hour += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district]["Total electricity consumption"][
                        "value"
                    ]
                    - self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "Total building electricity consumption"
                    ]["value"]
                )
                self_consumption_district_in_kilowatt_hour += self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Self-consumption of electricity"]["value"]

        overall_production_district_in_kilowatt_hour = (
            electricity_produktion_all_single_buildings_in_kilowatt_hour
            + electricity_production_district_in_kilowatt_hour
        )

        overall_consumption_district_in_kilowatt_hour = (
            electricity_consumption_all_single_buildings_in_kilowatt_hour
            + electricity_consumption_district_in_kilowatt_hour
        )

        overall_self_consumption_district_in_kilowatt_hour = (
            self_consumption_all_single_buildings_in_kilowatt_hour + self_consumption_district_in_kilowatt_hour
        )

        electricity_consumption_all_single_buildings_entry = KpiEntry(
            name="Electricity consumption of all single buildings",
            unit="kWh",
            value=electricity_consumption_all_single_buildings_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        electricity_production_all_single_buildings_entry = KpiEntry(
            name="Electricity production of all single buildings",
            unit="kWh",
            value=electricity_produktion_all_single_buildings_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        self_consumption_all_single_buildings_entry = KpiEntry(
            name="Self-consumption of all single buildings",
            unit="kWh",
            value=self_consumption_all_single_buildings_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        electricity_production_district_entry = KpiEntry(
            name="Electricity production of district without buildings",
            unit="kWh",
            value=electricity_production_district_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        electricity_consumption_district_entry = KpiEntry(
            name="Electricity consumption of district without buildings",
            unit="kWh",
            value=electricity_consumption_district_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        self_consumption_district_entry = KpiEntry(
            name="Self-consumption of district without buildings",
            unit="kWh",
            value=self_consumption_district_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        overall_production_district_entry = KpiEntry(
            name="Overall electricity production in district",
            unit="kWh",
            value=overall_production_district_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        overall_consumption_district_entry = KpiEntry(
            name="Overall electricity consumption in district",
            unit="kWh",
            value=overall_consumption_district_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )
        overall_self_consumption_district_entry = KpiEntry(
            name="Overall self-consumption in district",
            unit="kWh",
            value=overall_self_consumption_district_in_kilowatt_hour,
            tag=KpiTagEnumClass.GENERAL,
        )

        # update kpi collection dict
        self.kpi_collection_dict_unsorted[district_name].update(
            {
                electricity_consumption_all_single_buildings_entry.name: electricity_consumption_all_single_buildings_entry.to_dict(),
                electricity_production_all_single_buildings_entry.name: electricity_production_all_single_buildings_entry.to_dict(),
                self_consumption_all_single_buildings_entry.name: self_consumption_all_single_buildings_entry.to_dict(),
                electricity_production_district_entry.name: electricity_production_district_entry.to_dict(),
                electricity_consumption_district_entry.name: electricity_consumption_district_entry.to_dict(),
                self_consumption_district_entry.name: self_consumption_district_entry.to_dict(),
                overall_production_district_entry.name: overall_production_district_entry.to_dict(),
                overall_consumption_district_entry.name: overall_consumption_district_entry.to_dict(),
                overall_self_consumption_district_entry.name: overall_self_consumption_district_entry.to_dict(),
            }
        )

        return (
            overall_production_district_in_kilowatt_hour,
            overall_consumption_district_in_kilowatt_hour,
            overall_self_consumption_district_in_kilowatt_hour,
        )

    def create_overall_district_costs_collection(self, district_name):
        """Overall kpis for districts."""

        total_investment_cost_for_equipment_per_simulated_period_all_single_buildings = 0.0
        total_investment_cost_per_simulated_period_without_hp_all_single_buildings = 0.0
        total_investment_cost_per_simulated_period_only_hp_all_single_buildings = 0.0
        total_maintenance_cost_per_simulated_period_all_single_buildings = 0.0
        total_maintenance_cost_per_simulated_period_without_hp_all_single_buildings = 0.0
        total_maintenance_cost_per_simulated_period_only_hp_all_single_buildings = 0.0

        total_investment_cost_for_equipment_per_simulated_period_district = 0.0
        total_investment_cost_per_simulated_period_without_hp_district = 0.0
        total_investment_cost_per_simulated_period_only_hp_district = 0.0
        total_maintenance_cost_per_simulated_period_district = 0.0
        total_maintenance_cost_per_simulated_period_without_hp_district = 0.0
        total_maintenance_cost_per_simulated_period_only_hp_district = 0.0

        for building_objects_in_district in self.building_objects_in_district_list:
            if not any(word in building_objects_in_district for word in DistrictNames):
                total_investment_cost_for_equipment_per_simulated_period_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "Investment costs for equipment per simulated period"
                    ]["value"]
                )

                total_investment_cost_per_simulated_period_without_hp_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "Investment costs for equipment without heatpump per simulated period"
                    ]["value"]
                )

                total_investment_cost_per_simulated_period_only_hp_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "Investment costs for equipment only heatpump per simulated period"
                    ]["value"]
                )

                total_maintenance_cost_per_simulated_period_all_single_buildings += self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Maintenance costs for simulated period"]["value"]

                total_maintenance_cost_per_simulated_period_without_hp_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "Maintenance costs without heatpump for simulated period"
                    ]["value"]
                )

                total_maintenance_cost_per_simulated_period_only_hp_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "Maintenance costs only heatpump for simulated period"
                    ]["value"]
                )

            if any(word in building_objects_in_district for word in DistrictNames):
                total_investment_cost_for_equipment_per_simulated_period_district = self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Investment costs for equipment per simulated period"]["value"]

                total_investment_cost_per_simulated_period_without_hp_district = self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Investment costs for equipment without heatpump per simulated period"]["value"]

                total_investment_cost_per_simulated_period_only_hp_district = self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Investment costs for equipment only heatpump per simulated period"]["value"]

                total_maintenance_cost_per_simulated_period_district = self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Maintenance costs for simulated period"]["value"]

                total_maintenance_cost_per_simulated_period_without_hp_district = self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Maintenance costs without heatpump for simulated period"]["value"]

                total_maintenance_cost_per_simulated_period_only_hp_district = self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["Maintenance costs only heatpump for simulated period"]["value"]

        overall_investment_cost_for_equipment_per_simulated_period_district = (
            total_investment_cost_for_equipment_per_simulated_period_all_single_buildings
            + total_investment_cost_for_equipment_per_simulated_period_district
        )

        overall_investment_cost_per_simulated_period_without_hp_district = (
            total_investment_cost_per_simulated_period_without_hp_all_single_buildings
            + total_investment_cost_per_simulated_period_without_hp_district
        )

        overall_investment_cost_per_simulated_period_only_hp_district = (
            total_investment_cost_per_simulated_period_only_hp_all_single_buildings
            + total_investment_cost_per_simulated_period_only_hp_district
        )

        overall_maintenance_cost_per_simulated_period_district = (
            total_maintenance_cost_per_simulated_period_all_single_buildings
            + total_maintenance_cost_per_simulated_period_district
        )

        overall_maintenance_cost_per_simulated_period_without_hp_district = (
            total_maintenance_cost_per_simulated_period_without_hp_all_single_buildings
            + total_maintenance_cost_per_simulated_period_without_hp_district
        )

        overall_maintenance_cost_per_simulated_period_only_hp_district = (
            total_maintenance_cost_per_simulated_period_only_hp_all_single_buildings
            + total_maintenance_cost_per_simulated_period_only_hp_district
        )

        total_investment_cost_for_equipment_per_simulated_period_all_single_buildings_entry = KpiEntry(
            name="Total investment costs for equipment for all single buildings per simulated period",
            unit="EUR",
            value=total_investment_cost_for_equipment_per_simulated_period_all_single_buildings,
            tag=KpiTagEnumClass.COSTS,
        )
        total_investment_cost_per_simulated_period_without_hp_all_single_buildings_entry = KpiEntry(
            name="Total investment costs without heatpump for all single buildings per simulated period",
            unit="EUR",
            value=total_investment_cost_per_simulated_period_without_hp_all_single_buildings,
            tag=KpiTagEnumClass.COSTS,
        )
        total_investment_cost_per_simulated_period_only_hp_all_single_buildings_entry = KpiEntry(
            name="Total investment costs only heatpump for all single buildings per simulated period",
            unit="EUR",
            value=total_investment_cost_per_simulated_period_only_hp_all_single_buildings,
            tag=KpiTagEnumClass.COSTS,
        )
        total_maintenance_cost_per_simulated_period_all_single_buildings_entry = KpiEntry(
            name="Total maintenance costs for all single buildings per simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_all_single_buildings,
            tag=KpiTagEnumClass.COSTS,
        )
        total_maintenance_cost_per_simulated_period_without_hp_all_single_buildings_entry = KpiEntry(
            name="Total maintenance costs without heatpump for all single buildings per simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_without_hp_all_single_buildings,
            tag=KpiTagEnumClass.COSTS,
        )
        total_maintenance_cost_per_simulated_period_only_hp_all_single_buildings_entry = KpiEntry(
            name="Total maintenance costs only heatpump for all single buildings per simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_only_hp_all_single_buildings,
            tag=KpiTagEnumClass.COSTS,
        )
        total_investment_cost_for_equipment_per_simulated_period_district_entry = KpiEntry(
            name="Total investment costs for equipment for district without buildings per simulated period",
            unit="EUR",
            value=total_investment_cost_for_equipment_per_simulated_period_district,
            tag=KpiTagEnumClass.COSTS,
        )
        total_investment_cost_per_simulated_period_without_hp_district_entry = KpiEntry(
            name="Total investment costs without heatpump for district per simulated period",
            unit="EUR",
            value=total_investment_cost_per_simulated_period_without_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )
        total_investment_cost_per_simulated_period_only_hp_district_entry = KpiEntry(
            name="Total investment costs only heatpump for district per simulated period",
            unit="EUR",
            value=total_investment_cost_per_simulated_period_only_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )
        total_maintenance_cost_per_simulated_period_district_entry = KpiEntry(
            name="Total maintenance costs for district without buildings per simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_district,
            tag=KpiTagEnumClass.COSTS,
        )
        total_maintenance_cost_per_simulated_period_without_hp_district_entry = KpiEntry(
            name="Total maintenance costs without heatpump for district per simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_without_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )
        total_maintenance_cost_per_simulated_period_only_hp_district_entry = KpiEntry(
            name="Total maintenance costs only heatpump for district per simulated period",
            unit="EUR",
            value=total_maintenance_cost_per_simulated_period_only_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )
        overall_investment_cost_for_equipment_per_simulated_period_district_entry = KpiEntry(
            name="Overall investment for equipment costs in district per simulated period",
            unit="EUR",
            value=overall_investment_cost_for_equipment_per_simulated_period_district,
            tag=KpiTagEnumClass.COSTS,
        )
        overall_investment_cost_per_simulated_period_without_hp_district_entry = KpiEntry(
            name="Overall investment costs without heatpump in district per simulated period",
            unit="EUR",
            value=overall_investment_cost_per_simulated_period_without_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )
        overall_investment_cost_per_simulated_period_only_hp_district_entry = KpiEntry(
            name="Overall investment costs only heatpump in district per simulated period",
            unit="EUR",
            value=overall_investment_cost_per_simulated_period_only_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )
        overall_maintenance_cost_per_simulated_period_district_entry = KpiEntry(
            name="Overall maintenance costs in district per simulated period",
            unit="EUR",
            value=overall_maintenance_cost_per_simulated_period_district,
            tag=KpiTagEnumClass.COSTS,
        )
        overall_maintenance_cost_per_simulated_period_without_hp_district_entry = KpiEntry(
            name="Overall maintenance costs without heatpump in district per simulated period",
            unit="EUR",
            value=overall_maintenance_cost_per_simulated_period_without_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )
        overall_maintenance_cost_per_simulated_period_only_hp_district_entry = KpiEntry(
            name="Overall maintenance costs only heatpump in district per simulated period",
            unit="EUR",
            value=overall_maintenance_cost_per_simulated_period_only_hp_district,
            tag=KpiTagEnumClass.COSTS,
        )

        self.kpi_collection_dict_unsorted[district_name].update(
            {
                total_investment_cost_for_equipment_per_simulated_period_all_single_buildings_entry.name:
                    total_investment_cost_for_equipment_per_simulated_period_all_single_buildings_entry.to_dict(),
                total_investment_cost_per_simulated_period_without_hp_all_single_buildings_entry.name:
                    total_investment_cost_per_simulated_period_without_hp_all_single_buildings_entry.to_dict(),
                total_investment_cost_per_simulated_period_only_hp_all_single_buildings_entry.name:
                    total_investment_cost_per_simulated_period_only_hp_all_single_buildings_entry.to_dict(),
                total_maintenance_cost_per_simulated_period_all_single_buildings_entry.name:
                    total_maintenance_cost_per_simulated_period_all_single_buildings_entry.to_dict(),
                total_maintenance_cost_per_simulated_period_without_hp_all_single_buildings_entry.name:
                    total_maintenance_cost_per_simulated_period_without_hp_all_single_buildings_entry.to_dict(),
                total_maintenance_cost_per_simulated_period_only_hp_all_single_buildings_entry.name:
                    total_maintenance_cost_per_simulated_period_only_hp_all_single_buildings_entry.to_dict(),
                total_investment_cost_for_equipment_per_simulated_period_district_entry.name:
                    total_investment_cost_for_equipment_per_simulated_period_district_entry.to_dict(),
                total_investment_cost_per_simulated_period_without_hp_district_entry.name:
                    total_investment_cost_per_simulated_period_without_hp_district_entry.to_dict(),
                total_investment_cost_per_simulated_period_only_hp_district_entry.name:
                    total_investment_cost_per_simulated_period_only_hp_district_entry.to_dict(),
                total_maintenance_cost_per_simulated_period_district_entry.name:
                    total_maintenance_cost_per_simulated_period_district_entry.to_dict(),
                total_maintenance_cost_per_simulated_period_without_hp_district_entry.name:
                    total_maintenance_cost_per_simulated_period_without_hp_district_entry.to_dict(),
                total_maintenance_cost_per_simulated_period_only_hp_district_entry.name:
                    total_maintenance_cost_per_simulated_period_only_hp_district_entry.to_dict(),
                overall_investment_cost_for_equipment_per_simulated_period_district_entry.name:
                    overall_investment_cost_for_equipment_per_simulated_period_district_entry.to_dict(),
                overall_investment_cost_per_simulated_period_without_hp_district_entry.name:
                    overall_investment_cost_per_simulated_period_without_hp_district_entry.to_dict(),
                overall_investment_cost_per_simulated_period_only_hp_district_entry.name:
                    overall_investment_cost_per_simulated_period_only_hp_district_entry.to_dict(),
                overall_maintenance_cost_per_simulated_period_district_entry.name:
                    overall_maintenance_cost_per_simulated_period_district_entry.to_dict(),
                overall_maintenance_cost_per_simulated_period_without_hp_district_entry.name:
                    overall_maintenance_cost_per_simulated_period_without_hp_district_entry.to_dict(),
                overall_maintenance_cost_per_simulated_period_only_hp_district_entry.name:
                    overall_maintenance_cost_per_simulated_period_only_hp_district_entry.to_dict(),
            }
        )

    def create_overall_district_emissions_collection(self, district_name):
        """Overall kpis for districts."""

        total_co2_emissions_for_equipment_per_simulated_period_all_single_buildings = 0.0
        total_co2_emissions_for_equipment_per_simulated_period_without_hp_all_single_buildings = 0.0
        total_co2_emissions_for_equipment_per_simulated_period_only_hp_all_single_buildings = 0.0

        total_co2_emissions_for_equipment_per_simulated_period_district = 0.0
        total_co2_emissions_for_equipment_per_simulated_period_without_hp_district = 0.0
        total_co2_emissions_for_equipment_per_simulated_period_only_hp_district = 0.0

        for building_objects_in_district in self.building_objects_in_district_list:
            if not any(word in building_objects_in_district for word in DistrictNames):
                total_co2_emissions_for_equipment_per_simulated_period_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "CO2 footprint for equipment per simulated period"
                    ]["value"]
                )

                total_co2_emissions_for_equipment_per_simulated_period_without_hp_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "CO2 footprint for equipment without heatpump per simulated period"
                    ]["value"]
                )

                total_co2_emissions_for_equipment_per_simulated_period_only_hp_all_single_buildings += (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "CO2 footprint for equipment only heatpump per simulated period"
                    ]["value"]
                )

            if any(word in building_objects_in_district for word in DistrictNames):
                total_co2_emissions_for_equipment_per_simulated_period_district = self.kpi_collection_dict_unsorted[
                    building_objects_in_district
                ]["CO2 footprint for equipment per simulated period"]["value"]

                total_co2_emissions_for_equipment_per_simulated_period_without_hp_district = (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "CO2 footprint for equipment without heatpump per simulated period"
                    ]["value"]
                )

                total_co2_emissions_for_equipment_per_simulated_period_only_hp_district = (
                    self.kpi_collection_dict_unsorted[building_objects_in_district][
                        "CO2 footprint for equipment only heatpump per simulated period"
                    ]["value"]
                )

        overall_co2_emission_for_equipment_per_simulated_period_district = (
            total_co2_emissions_for_equipment_per_simulated_period_all_single_buildings
            + total_co2_emissions_for_equipment_per_simulated_period_district
        )

        overall_co2_emission_per_simulated_period_without_hp_district = (
            total_co2_emissions_for_equipment_per_simulated_period_without_hp_all_single_buildings
            + total_co2_emissions_for_equipment_per_simulated_period_without_hp_district
        )

        overall_co2_emission_per_simulated_period_only_hp_district = (
            total_co2_emissions_for_equipment_per_simulated_period_only_hp_all_single_buildings
            + total_co2_emissions_for_equipment_per_simulated_period_only_hp_district
        )

        total_co2_emissions_per_simulated_period_all_single_buildings_entry = KpiEntry(
            name="CO2 footprint for equipment for all single buildings per simulated period",
            unit="kg",
            value=total_co2_emissions_for_equipment_per_simulated_period_all_single_buildings,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        total_co2_emissions_per_simulated_period_without_hp_all_single_buildings_entry = KpiEntry(
            name="CO2 footprint for equipment without heatpump for all single buildings per simulated period",
            unit="kg",
            value=total_co2_emissions_for_equipment_per_simulated_period_without_hp_all_single_buildings,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        total_co2_emissions_per_simulated_period_only_hp_all_single_buildings_entry = KpiEntry(
            name="CO2 footprint for equipment only heatpump for all single buildings per simulated period",
            unit="kg",
            value=total_co2_emissions_for_equipment_per_simulated_period_only_hp_all_single_buildings,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        total_co2_emissions_for_equipment_per_simulated_period_district_entry = KpiEntry(
            name="CO2 footprint for equipment for district without buildings per simulated period",
            unit="kg",
            value=total_co2_emissions_for_equipment_per_simulated_period_district,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        total_co2_emissions_per_simulated_period_without_hp_district_entry = KpiEntry(
            name="CO2 footprint for equipment without heatpump for district per simulated period",
            unit="kg",
            value=total_co2_emissions_for_equipment_per_simulated_period_without_hp_district,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        total_co2_emissions_per_simulated_period_only_hp_district_entry = KpiEntry(
            name="CO2 footprint for equipment only heatpump for district per simulated period",
            unit="kg",
            value=total_co2_emissions_for_equipment_per_simulated_period_only_hp_district,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        overall_co2_emissions_for_equipment_per_simulated_period_district_entry = KpiEntry(
            name="Overall CO2 footprint for equipment in district per simulated period",
            unit="kg",
            value=overall_co2_emission_for_equipment_per_simulated_period_district,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        overall_co2_emissions_per_simulated_period_without_hp_district_entry = KpiEntry(
            name="Overall CO2 footprint for equipment without heatpump in district per simulated period",
            unit="kg",
            value=overall_co2_emission_per_simulated_period_without_hp_district,
            tag=KpiTagEnumClass.EMISSIONS,
        )
        overall_co2_emissions_per_simulated_period_only_hp_district_entry = KpiEntry(
            name="Overall CO2 footprint for equipment only heatpump in district per simulated period",
            unit="kg",
            value=overall_co2_emission_per_simulated_period_only_hp_district,
            tag=KpiTagEnumClass.EMISSIONS,
        )

        self.kpi_collection_dict_unsorted[district_name].update(
            {
                total_co2_emissions_per_simulated_period_all_single_buildings_entry.name:
                    total_co2_emissions_per_simulated_period_all_single_buildings_entry.to_dict(),
                total_co2_emissions_per_simulated_period_without_hp_all_single_buildings_entry.name:
                    total_co2_emissions_per_simulated_period_without_hp_all_single_buildings_entry.to_dict(),
                total_co2_emissions_per_simulated_period_only_hp_all_single_buildings_entry.name:
                    total_co2_emissions_per_simulated_period_only_hp_all_single_buildings_entry.to_dict(),
                total_co2_emissions_for_equipment_per_simulated_period_district_entry.name:
                    total_co2_emissions_for_equipment_per_simulated_period_district_entry.to_dict(),
                total_co2_emissions_per_simulated_period_without_hp_district_entry.name:
                    total_co2_emissions_per_simulated_period_without_hp_district_entry.to_dict(),
                total_co2_emissions_per_simulated_period_only_hp_district_entry.name:
                    total_co2_emissions_per_simulated_period_only_hp_district_entry.to_dict(),
                overall_co2_emissions_for_equipment_per_simulated_period_district_entry.name:
                    overall_co2_emissions_for_equipment_per_simulated_period_district_entry.to_dict(),
                overall_co2_emissions_per_simulated_period_without_hp_district_entry.name:
                    overall_co2_emissions_per_simulated_period_without_hp_district_entry.to_dict(),
                overall_co2_emissions_per_simulated_period_only_hp_district_entry.name:
                    overall_co2_emissions_per_simulated_period_only_hp_district_entry.to_dict(),
            }
        )

    def create_overall_district_contracting_collection(self, district_name, all_outputs):
        """If buildings are in contracting with district, heatpump costs and emissions are on district side."""

        set_of_buildings_in_contracting = set()
        for building_objekt in self.building_objects_in_district_list:
            for output in all_outputs:
                if building_objekt == output.component_name.split("_")[0] and district_name in output.component_name:
                    set_of_buildings_in_contracting.add(building_objekt)

        if district_name in set_of_buildings_in_contracting:
            set_of_buildings_in_contracting.remove(district_name)
        list_of_buildings_in_contracting = list(set_of_buildings_in_contracting)

        if len(list_of_buildings_in_contracting) == 0:
            pass
        else:
            number_of_building_in_contracting = len(list_of_buildings_in_contracting)

            total_co2_emissions_for_hp_of_building_per_simulated_period_contracting = 0.0
            total_investment_cost_for_hp_of_building_per_simulated_period_contracting = 0.0
            total_maintenance_cost_for_hp_of_building_per_simulated_period_contracting = 0.0

            for building in list_of_buildings_in_contracting:
                total_co2_emissions_for_hp_of_building_per_simulated_period_contracting += (
                    self.kpi_collection_dict_unsorted[building][
                        "CO2 footprint for equipment only heatpump per simulated period"
                    ]["value"]
                )
                total_investment_cost_for_hp_of_building_per_simulated_period_contracting += (
                    self.kpi_collection_dict_unsorted[building][
                        "Investment costs for equipment only heatpump per simulated period"
                    ]["value"]
                )
                total_maintenance_cost_for_hp_of_building_per_simulated_period_contracting += (
                    self.kpi_collection_dict_unsorted[building][
                        "Maintenance costs only heatpump for simulated period"
                    ]["value"]
                )

            total_co2_emissions_of_district_grid_per_simulated_period_contracting = (
                self.kpi_collection_dict_unsorted[district_name][
                    "Total CO2 emissions for simulated period"
                ]["value"]
            )

            total_costs_of_district_grid_per_simulated_period_contracting = (
                self.kpi_collection_dict_unsorted[district_name][
                    "Total costs for simulated period"
                ]["value"]
            )

            overall_district_cost_contracting = \
                (total_costs_of_district_grid_per_simulated_period_contracting +
                 total_investment_cost_for_hp_of_building_per_simulated_period_contracting +
                 total_maintenance_cost_for_hp_of_building_per_simulated_period_contracting)

            overall_district_co2_emissions_contracting = \
                (total_co2_emissions_of_district_grid_per_simulated_period_contracting +
                 total_co2_emissions_for_hp_of_building_per_simulated_period_contracting)

            number_of_building_in_contracting_entry = KpiEntry(
                name="Number of buildings in district contracting",
                unit="-",
                value=number_of_building_in_contracting,
                tag=KpiTagEnumClass.CONTRACTING,
            )

            total_co2_emissions_for_hp_of_building_in_contracting_entry = KpiEntry(
                name="CO2 footprint of hp only of buildings in contracting",
                unit="kg",
                value=total_co2_emissions_for_hp_of_building_per_simulated_period_contracting,
                tag=KpiTagEnumClass.CONTRACTING,
            )

            total_investment_cost_for_hp_of_building_in_contracting_entry = KpiEntry(
                name="Investment cost of hp only of buildings in contracting",
                unit="EUR",
                value=total_investment_cost_for_hp_of_building_per_simulated_period_contracting,
                tag=KpiTagEnumClass.CONTRACTING,
            )

            total_maintenance_cost_for_hp_of_building_contracting_entry = KpiEntry(
                name="Maintenance cost of hp only of buildings in contracting",
                unit="EUR",
                value=total_maintenance_cost_for_hp_of_building_per_simulated_period_contracting,
                tag=KpiTagEnumClass.CONTRACTING,
            )
            overall_district_cost_contracting_entry = KpiEntry(
                name="Overall costs for grid and hp in contracting per simulated period",
                unit="EUR",
                value=overall_district_cost_contracting,
                tag=KpiTagEnumClass.CONTRACTING,
            )
            overall_district_co2_emissions_contracting_entry = KpiEntry(
                name="Overall CO2 footprint for grid and hp in contracting per simulated period",
                unit="kg",
                value=overall_district_co2_emissions_contracting,
                tag=KpiTagEnumClass.CONTRACTING,
            )

            self.kpi_collection_dict_unsorted[district_name].update(
                {
                    number_of_building_in_contracting_entry.name:
                        number_of_building_in_contracting_entry.to_dict(),
                    total_co2_emissions_for_hp_of_building_in_contracting_entry.name:
                        total_co2_emissions_for_hp_of_building_in_contracting_entry.to_dict(),
                    total_investment_cost_for_hp_of_building_in_contracting_entry.name:
                        total_investment_cost_for_hp_of_building_in_contracting_entry.to_dict(),
                    total_maintenance_cost_for_hp_of_building_contracting_entry.name:
                        total_maintenance_cost_for_hp_of_building_contracting_entry.to_dict(),
                    overall_district_cost_contracting_entry.name:
                        overall_district_cost_contracting_entry.to_dict(),
                    overall_district_co2_emissions_contracting_entry.name:
                        overall_district_co2_emissions_contracting_entry.to_dict(),
                }
            )

    def get_all_component_kpis(self, wrapped_components: List[ComponentWrapper]) -> None:
        """Go through all components and get their KPIs if implemented."""
        my_component_kpi_entry_list: List[KpiEntry]

        self.kpi_collection_dict_unsorted = {
            building_objects: {} for building_objects in self.building_objects_in_district_list
        }

        for wrapped_component in wrapped_components:
            my_component = wrapped_component.my_component
            # get KPIs of respective component
            my_component_kpi_entry_list = my_component.get_component_kpi_entries(
                all_outputs=self.all_outputs, postprocessing_results=self.results
            )

            if my_component_kpi_entry_list != []:
                # add all KPI entries to kpi dict
                for kpi_entry in my_component_kpi_entry_list:

                    for object_name in self.kpi_collection_dict_unsorted.keys():
                        if (object_name == my_component.component_name.split("_")[0] or
                                not self.simulation_parameters.multiple_buildings):
                            self.kpi_collection_dict_unsorted[object_name][kpi_entry.name] = kpi_entry.to_dict()
                            break
            else:
                log.debug(
                    "KPI generation for "
                    + my_component.component_name
                    + " was not successful. KPI method is maybe not implemented yet."
                )



# -*- coding: utf-8 -*-
# clean
""" Generic heating controller with configuration and state.

It controls the heating system (heat transfer from buffer storage to building)
only during the heating period.
It is a ping pong control with an optional input from the Energy Management System,
which enforces heating with electricity from PV.
The buffer is controlled accoring to four modes:
    (a) 0.5 * power when buffer temperature is within the upper half between upper target and increased upper target from Energy Management System (only in surplus case),
    (b) 0.75 * power when buffer temperature is within the lower half beweet upper target and increase upper target from Energy Management System (only in surplus case),
    (c) full power when building temperature is below lower target,
    (d) off when temperature is higher than upper target.
"""
import importlib
from dataclasses import dataclass

# Owned
from typing import List, Any

from dataclasses_json import dataclass_json

from hisim import utils
from hisim import component as cp
from hisim.components.building import Building
from hisim.components import controller_l2_energy_management_system
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class L1BuildingHeatingConfig(cp.ConfigBase):
    """Configuration of Building Controller."""

    building_name: str
    #: name of the device
    name: str
    #: priority of the device in hierachy: the higher the number the lower the priority
    source_weight: int
    #: lower set temperature of building, given in °C
    t_min_heating_in_celsius: float
    #: upper set temperature of building, given in °C
    t_max_heating_in_celsius: float
    #: upper temperature of buffer, where heating of building is enforced.
    t_buffer_activation_threshold_in_celsius: float
    # julian day of simulation year, where heating season begins
    day_of_heating_season_begin: int
    # julian day of simulation year, where heating season ends
    day_of_heating_season_end: int

    @staticmethod
    def get_default_config_heating(
        name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Default config for the heating controller."""
        config = L1BuildingHeatingConfig(
            building_name=building_name,
            name="L1BuildingTemperatureController" + name,
            source_weight=1,
            t_min_heating_in_celsius=19.5,
            t_max_heating_in_celsius=20.5,
            t_buffer_activation_threshold_in_celsius=40.0,
            day_of_heating_season_begin=270,
            day_of_heating_season_end=150,
        )
        return config


class L1BuildingHeatControllerState:
    """Data class that saves the state of the controller."""

    def __init__(self, state: float = 0):
        """Initializes the class."""
        self.state: float = state

    def clone(self) -> "L1BuildingHeatControllerState":
        """Clones itself."""
        return L1BuildingHeatControllerState(state=self.state)


class L1BuildingHeatController(cp.Component):
    """L1 building controller. Processes signals ensuring comfort temperature of building.

    Gets temperature of building to control as input, as well as a signal from the energy management system to increase the set temperatur of the buffer storage.
    It outputs a control signal with four modes (0, 0.5, 0.75 and 1) for zero, half, three quarter and full power accordingly.
    It is only activated during the heating season.

    Components to connect to:
    (1) Buffer (generic_hot_water_storage_modular)
    (2) Building (building)
    (3) Energy Management System (controller_l2_energy_management_system) - optional
    """

    # Inputs
    BuildingTemperature = "BuildingTemperature"
    BuildingTemperatureModifier = "BuildingTemperatureModifier"
    BufferTemperature = "BufferTemperature"
    # Outputs
    HeatControllerTargetPercentage = "HeatControllerTargetPercentage"

    # #Forecasts
    # HeatPumpLoadForecast = "HeatPumpLoadForecast"

    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: L1BuildingHeatingConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """For initializing."""
        if not config.__class__.__name__ == L1BuildingHeatingConfig.__name__:
            raise ValueError("Wrong config class.")
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.config: L1BuildingHeatingConfig = config

        """ Initializes the class. """
        self.source_weight: int = config.source_weight
        self.heating_season_begin = (
            config.day_of_heating_season_begin * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
        )
        self.heating_season_end = (
            config.day_of_heating_season_end * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
        )
        self.state: L1BuildingHeatControllerState = L1BuildingHeatControllerState()
        self.previous_state: L1BuildingHeatControllerState = L1BuildingHeatControllerState()
        self.processed_state: L1BuildingHeatControllerState = L1BuildingHeatControllerState()

        # Component Outputs
        self.heat_controller_target_percentage_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.HeatControllerTargetPercentage,
            LoadTypes.ON_OFF,
            Units.BINARY,
            output_description="Heating controller of buffer storage.",
        )

        # Component Inputs
        self.building_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.BuildingTemperature,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            mandatory=True,
        )

        self.buffer_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.BufferTemperature,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            mandatory=False,
        )

        self.building_temperature_modifier_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.BuildingTemperatureModifier,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            mandatory=False,
        )

        self.add_default_connections(self.get_building_default_connections())
        self.add_default_connections(self.get_default_connections_from_hot_water_storage())
        self.add_default_connections(self.get_default_connections_from_ems())

    def get_building_default_connections(self):
        """Sets the default connections for the building."""

        connections = []
        building_classname = Building.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1BuildingHeatController.BuildingTemperature,
                building_classname,
                Building.TemperatureMeanThermalMass,
            )
        )
        return connections

    def get_default_connections_from_ems(self):
        """Sets the default connections for the energy management system."""

        connections = []
        ems_classname = controller_l2_energy_management_system.L2GenericEnergyManagementSystem.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1BuildingHeatController.BuildingTemperatureModifier,
                ems_classname,
                controller_l2_energy_management_system.L2GenericEnergyManagementSystem.BuildingIndoorTemperatureModifier,
            )
        )
        return connections

    def get_default_connections_from_hot_water_storage(self):
        """Sets default connections for the buffer."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.generic_hot_water_storage_modular"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "HotWaterStorage")
        connections = []
        boiler_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1BuildingHeatController.BufferTemperature,
                boiler_classname,
                component_class.TemperatureMean,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def control_heating(
        self,
        timestep: int,
        t_control: float,
        t_buffer: float,
        temperature_modifier: float,
    ) -> None:
        """Controls the heating from buffer to building."""
        # prevent heating in summer
        if self.heating_season_begin > timestep > self.heating_season_end:
            self.state.state = 0
            return
        # activate heating when building temperature is below lower threshold
        if t_control < self.config.t_min_heating_in_celsius:
            # start heating if temperature goes below lower limit
            self.state.state = 1
            return
        # deactivate heating when building temperature is above upper threshold
        if t_control > self.config.t_max_heating_in_celsius + temperature_modifier:
            self.state.state = 0
            return
        # deactivate heating when temperature modifier is zero and signal comes from surplus control.
        # states 0.5 and 0.75 are only activated when temperature modifier is greater than zero, which is only the case in surplus control.
        if self.state.state in [0.5, 0.75] and temperature_modifier == 0:
            self.state.state = 0
            return
        # "surplus heat control" when storage is getting hot
        if temperature_modifier > 0 and t_buffer > self.config.t_buffer_activation_threshold_in_celsius:
            # heat with 75 % power and building can still be heated
            if t_control < self.config.t_max_heating_in_celsius + temperature_modifier / 2:
                self.state.state = 0.75
            # heat with 50 % power when storage is getting hot and building can still be heated, but is already on the upper side of the tolerance interval
            elif t_control < self.config.t_max_heating_in_celsius + temperature_modifier:
                self.state.state = 0.5
        return

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores previous state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """For double checking results."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the control of the building temperature, when building is heated from buffer."""
        if force_convergence:
            pass
        else:
            # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
            t_control = stsv.get_input_value(self.building_temperature_channel)
            if self.buffer_temperature_channel.source_output is not None:
                t_buffer = stsv.get_input_value(self.buffer_temperature_channel)
            else:
                t_buffer = 0
            temperature_modifier = stsv.get_input_value(self.building_temperature_modifier_channel)
            self.control_heating(
                timestep=timestep,
                t_control=t_control,
                t_buffer=t_buffer,
                temperature_modifier=temperature_modifier,
            )
            self.processed_state = self.state.clone()
        stsv.set_output_value(self.heat_controller_target_percentage_channel, self.processed_state.state)

    def write_to_report(self) -> List[str]:
        """Writes the information of the current component to the report."""
        return self.config.get_string_dict()



""" Simple hot water storage implementation: HotWaterStorage class together with state and configuration class.

Energy bucket model: extracts energy, adds energy and converts back to temperatere.
The hot water storage simulates only storage and demand and needs to be connnected to a heat source. It can act as
DHW hot water storage or as buffer storage.
"""

import importlib
from dataclasses import dataclass

# clean
# Generic/Built-in
from typing import Any, List, Optional

import numpy as np
import pandas as pd
from dataclasses_json import dataclass_json

# Owned
import hisim.component as cp
import hisim.log
from hisim import loadtypes as lt
from hisim.component import OpexCostDataClass, CapexCostDataClass
from hisim.components import (
    controller_l1_building_heating,
    generic_chp,
    configuration,
)
from hisim.components.loadprofilegenerator_utsp_connector import UtspLpgConnector
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiTagEnumClass, KpiEntry

__authors__ = "Johanna Ganglbauer - johanna.ganglbauer@4wardenergy.at"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class StorageConfig(cp.ConfigBase):
    """Used in the HotWaterStorageClass defining the basics."""

    building_name: str
    #: name of the hot water storage
    name: str
    #: priority of the component in hierachy: the higher the number the lower the priority
    source_weight: int
    #: type of use, either boiler or buffer
    use: lt.ComponentType
    #: volume of storage in m^3
    volume: float
    #: surface of storage in m^2
    surface: float
    #: u-value of storage in W/(K m^2)
    u_value: float
    #: energy of full cycle in kWh
    energy_full_cycle: Optional[float]
    #: power of heat source in kW
    power: float
    #: CO2 footprint of investment in kg
    co2_footprint: float
    #: cost for investment in Euro
    cost: float
    #: lifetime in years
    lifetime: float
    # maintenance cost as share of investment [0..1]
    maintenance_cost_as_percentage_of_investment: float

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return HotWaterStorage.get_full_classname()

    @classmethod
    def get_default_config_for_boiler(
        cls,
        name: str = "DHWBoiler",
        building_name: str = "BUI1",
    ) -> "StorageConfig":
        """Returns default configuration for boiler."""

        volume = 250
        radius = (volume * 1e-3 / (4 * np.pi)) ** (1 / 3)  # l to m^3 so that radius is given in m
        surface = 2 * radius * radius * np.pi + 2 * radius * np.pi * (4 * radius)
        config = StorageConfig(
            building_name=building_name,
            name=name,
            use=lt.ComponentType.BOILER,
            source_weight=1,
            volume=volume,
            surface=surface,
            u_value=0.36,
            energy_full_cycle=None,
            power=0,
            co2_footprint=0,  # Todo: check value
            cost=volume * 14.51,  # value from emission_factros_and_costs_devices.csv
            lifetime=20,  # SOURCE: VDI2067-1
            maintenance_cost_as_percentage_of_investment=0.02,  # SOURCE: VDI2067-1
        )
        return config

    @classmethod
    def get_scaled_config_for_boiler_to_number_of_apartments(
        cls,
        number_of_apartments: float,
        default_volume_in_liter: float = 250.0,
        name: str = "DHWBoiler",
        building_name: str = "BUI1",
    ) -> "StorageConfig":
        """Returns default configuration for boiler."""

        volume = default_volume_in_liter * max(number_of_apartments, 1)
        radius = (volume * 1e-3 / (4 * np.pi)) ** (1 / 3)  # l to m^3 so that radius is given in m
        surface = 2 * radius * radius * np.pi + 2 * radius * np.pi * (4 * radius)
        config = StorageConfig(
            building_name=building_name,
            name=name,
            use=lt.ComponentType.BOILER,
            source_weight=1,
            volume=volume,
            surface=surface,
            u_value=0.36,
            energy_full_cycle=None,
            power=0,
            co2_footprint=0,  # Todo: check value
            cost=volume * 14.51,  # value from emission_factros_and_costs_devices.csv
            lifetime=20,  # SOURCE: VDI2067-1
            maintenance_cost_as_percentage_of_investment=0.02,  # SOURCE: VDI2067-1
        )
        return config

    @staticmethod
    def get_default_config_buffer(
        name: str = "Buffer",
        power: float = 2000,
        volume: float = 500,
        building_name: str = "BUI1",
    ) -> Any:
        """Returns default configuration for buffer (radius:height = 1:4)."""
        # volume = r^2 * pi * h = r^2 * pi * 4r = 4 * r^3 * pi
        radius = (volume * 1e-3 / (4 * np.pi)) ** (1 / 3)  # l to m^3 so that radius is given in m
        # cylinder surface area = floor and ceiling area + lateral surface
        surface = 2 * radius * radius * np.pi + 2 * radius * np.pi * (4 * radius)
        config = StorageConfig(
            building_name=building_name,
            name=name,
            use=lt.ComponentType.BUFFER,
            source_weight=1,
            volume=0,
            surface=surface,
            u_value=0.36,
            energy_full_cycle=None,
            power=power,
            co2_footprint=100,  # Todo: check value
            cost=volume * 14.51,  # value from emission_factros_and_costs_devices.csv
            lifetime=100,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.02,  # SOURCE: VDI2067-1
        )
        return config

    def compute_default_volume(
        self,
        time_in_seconds: float,
        temperature_difference_in_kelvin: float,
        multiplier: float,
    ) -> None:
        """Computes default volume and surface from power and min idle time of heating system."""
        if self.use != lt.ComponentType.BUFFER:
            raise Exception("Default volume can only be computed for buffer storage not for boiler.")

        energy_in_kilo_joule = self.power * time_in_seconds * 1e-3
        self.volume = energy_in_kilo_joule * multiplier / (temperature_difference_in_kelvin * 0.977 * 4.182)
        # volume = r^2 * pi * h = r^2 * pi * 4r = 4 * r^3 * pi
        radius = (self.volume * 1e-3 / (4 * np.pi)) ** (1 / 3)  # l to m^3 so that radius is given in m
        # cylinder surface area = floor and ceiling area + lateral surface
        self.surface = 2 * radius * radius * np.pi + 2 * radius * np.pi * (4 * radius)

    def compute_default_cycle(self, temperature_difference_in_kelvin: float) -> None:
        """Computes energy needed to heat storage from lower threshold of hysteresis to upper threshold."""
        self.energy_full_cycle = self.volume * temperature_difference_in_kelvin * 0.977 * 4.182 / 3600


class StorageState:
    """Data class saves the state of the simulation results."""

    def __init__(
        self,
        timestep: int = -1,
        volume_in_l: float = 200,
        temperature_in_kelvin: float = 273.15 + 50,
    ):
        """Initializes instance of class Storage State.

        :param timestep: timestep of simulation
        :type timestep: int, optional
        :param volume_in_l: volume of hot water storage in liters
        :type volume_in_l: float
        :param temperature_in_kelvin: temperature of hot water storage in Kelvin
        :type temperature_in_kelvin: float
        """
        self.timestep = timestep
        self.temperature_in_kelvin = temperature_in_kelvin
        self.volume_in_l = volume_in_l

    def clone(self):
        """Replicates storage state."""
        return StorageState(self.timestep, self.volume_in_l, self.temperature_in_kelvin)

    def energy_from_temperature(self) -> float:
        """Converts temperature of storage (K) into energy contained in storage (kJ)."""
        # 0.977 is the density of water in kg/l
        # 4.182 is the specific heat of water in kJ / (K * kg)
        return self.temperature_in_kelvin * self.volume_in_l * 0.977 * 4.182  # energy given in kJ

    def set_temperature_from_energy(self, energy_in_kilo_joule: float) -> None:
        """Converts energy contained in storage (kJ) into temperature (K)."""
        # 0.977 is the density of water in kg/l
        # 4.182 is the specific heat of water in kJ / (K * kg)
        self.temperature_in_kelvin = energy_in_kilo_joule / (self.volume_in_l * 0.977 * 4.182)  # temperature given in K
        # filter for boiling water
        # no filtering -> this hides major problems - Noah
        if self.temperature_in_kelvin > 95 + 273.15:
            raise ValueError(
                "Water was boiling. This points towards a major problem in your model. Increasing the storage volume may solve the issue"
            )
        # filter for freezing water
        if self.temperature_in_kelvin < 2 + 273.15:
            raise ValueError(
                "Water in your storage tank was freezing. This points towards a major problem in your model."
            )

    def return_available_energy(self) -> float:
        """Returns available energy in (kJ).

        For heating up the building in winter. Here 30°C is set as the lower limit for the temperature in the buffer storage in winter.
        """
        return (self.temperature_in_kelvin - 273.15 - 25) * self.volume_in_l * 0.977 * 4.182


# class HotWaterStorage(dycp.DynamicComponent):
class HotWaterStorage(cp.Component):
    """Simple hot water storage implementation.

    Energy bucket model: extracts energy, adds energy and converts back to temperatere.
    The hot water storage simulates only storage and demand and needs to be connnected to a heat source. It can act as boiler with input:
    WaterConsumption or as  buffer with input ThermalPowerDelivered from building component. Both options need input signal for heating power and have
    two outputs: the hot water storage temperature, and the power extracted from the hot water storage.

    Components to connect to:
    (1a) Building Controller(controller_l1_building_heating) - if buffer
    (1b) Occupancy Profile (either loadprofilegenerator_connector or loadprofilegenerator_utsp_connector) - if boiler
    (2a) Heat Source (generic_heat_source)
    (2b) Heat Pump (generic_heat_pump_modular)
    (2c) CHP (generic_CHP) - optional - if CHP additionally heats bufffer
    """

    # Inputs
    ThermalPowerDelivered = "ThermalPowerDelivered"
    ThermalPowerCHP = "ThermalPowerCHP"
    WaterConsumption = "WaterConsumption"
    HeatControllerTargetPercentage = "HeatControllerTargetPercentage"
    # my_component_inputs: List[dycp.DynamicConnectionInput] = []
    # my_component_outputs: List[dycp.DynamicConnectionOutput] = []

    # obligatory Outputs
    TemperatureMean = "TemperatureMean"

    # outputs for buffer storage
    PowerFromHotWaterStorage = "PowerFromHotWaterStorage"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: StorageConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(display_in_webtool=True),
    ):
        """Initializes instance of HotWaterStorage class."""

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.build(config)

        # collect all heat inputs
        self.heat_to_buffer_inputs: List[cp.ComponentInput]

        # initialize Boiler State
        self.state = StorageState(volume_in_l=config.volume, temperature_in_kelvin=273 + 60)
        self.previous_state = self.state.clone()
        self.write_to_report()

        # inputs
        if self.use == lt.ComponentType.BOILER:
            self.water_consumption_channel: cp.ComponentInput = self.add_input(
                self.component_name,
                self.WaterConsumption,
                lt.LoadTypes.WARM_WATER,
                lt.Units.LITER,
                mandatory=True,
            )
            self.add_default_connections(self.get_utsp_default_connections())
        elif self.use == lt.ComponentType.BUFFER:
            self.heat_controller_target_percentage_channel: cp.ComponentInput = self.add_input(
                self.component_name,
                self.HeatControllerTargetPercentage,
                lt.LoadTypes.ON_OFF,
                lt.Units.BINARY,
                mandatory=True,
            )
            self.add_default_connections(self.get_heating_controller_default_connections())
        else:
            hisim.log.error("Type of hot water storage is not defined")

        self.thermal_power_delivered_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ThermalPowerDelivered,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            mandatory=False,
        )
        self.thermal_power_chp_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ThermalPowerCHP,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            mandatory=False,
        )

        # Outputs
        self.temperature_mean_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TemperatureMean,
            load_type=lt.LoadTypes.TEMPERATURE,
            unit=lt.Units.CELSIUS,
            postprocessing_flag=[lt.InandOutputType.STORAGE_CONTENT],
            output_description="Temperature Mean",
        )
        # Outputs
        self.power_from_water_storage_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.PowerFromHotWaterStorage,
            lt.LoadTypes.HEATING,
            lt.Units.KILOJOULE,
            postprocessing_flag=[
                lt.InandOutputType.DISCHARGE,
                self.use,
                lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
            ],
            output_description="Power transfered to Building or Hot Water Pipe.",
        )

        self.add_default_connections(self.get_default_connections_from_generic_heat_pump_modular())
        self.add_default_connections(self.get_heatsource_default_connections())
        self.add_default_connections(self.get_chp_default_connections())

    def get_utsp_default_connections(self):
        """Sets occupancy default connections in hot water storage."""

        connections = []
        utsp_classname = UtspLpgConnector.get_classname()
        connections.append(
            cp.ComponentConnection(
                HotWaterStorage.WaterConsumption,
                utsp_classname,
                UtspLpgConnector.WaterConsumption,
            )
        )
        return connections

    def get_default_connections_from_generic_heat_pump_modular(self):
        """Sets heat pump default connections in hot water storage."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.generic_heat_pump_modular"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "ModularHeatPump")

        connections = []
        heatpump_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                HotWaterStorage.ThermalPowerDelivered,
                heatpump_classname,
                component_class.ThermalPowerDelivered,
            )
        )
        return connections

    def get_heatsource_default_connections(self):
        """Sets heat source default connections in hot water storage."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.generic_heat_source"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "HeatSource")

        connections = []
        heatsource_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                HotWaterStorage.ThermalPowerDelivered,
                heatsource_classname,
                component_class.ThermalPowerDelivered,
            )
        )
        return connections

    def get_chp_default_connections(self):
        """Sets chp default connections in hot water storage."""

        connections = []
        chp_classname = generic_chp.SimpleCHP.get_classname()
        connections.append(
            cp.ComponentConnection(
                HotWaterStorage.ThermalPowerCHP,
                chp_classname,
                generic_chp.SimpleCHP.ThermalPowerOutputBoiler,
            )
        )
        return connections

    def get_heating_controller_default_connections(self):
        """Sets heating controller default connections in hot water storage."""

        connections = []
        heating_controller_classname = controller_l1_building_heating.L1BuildingHeatController.get_classname()
        connections.append(
            cp.ComponentConnection(
                HotWaterStorage.HeatControllerTargetPercentage,
                heating_controller_classname,
                controller_l1_building_heating.L1BuildingHeatController.HeatControllerTargetPercentage,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def build(self, config: StorageConfig) -> None:
        """Initializes hot water storage instance."""

        self.name = config.name
        self.use = config.use
        self.source_weight = config.source_weight
        self.volume = config.volume
        self.surface = config.surface
        self.u_value = config.u_value
        self.drain_water_temperature = configuration.HouseholdWarmWaterDemandConfig.freshwater_temperature
        self.warm_water_temperature = (
            configuration.HouseholdWarmWaterDemandConfig.ww_temperature_demand
            - configuration.HouseholdWarmWaterDemandConfig.temperature_difference_hot
        )
        self.power = config.power
        self.config = config

    def write_to_report(self):
        """Writes to report."""
        return self.config.get_string_dict()

    def i_save_state(self):
        """Abstract. Gets called at the beginning of a timestep to save the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self):
        """Abstract. Restores the state of the component. Can be called many times while iterating."""
        self.state = self.previous_state.clone()

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates iteration of hot water storage."""

        thermal_energy_delivered = 0.0
        if self.thermal_power_delivered_channel.source_output is not None:
            thermal_energy_delivered = (
                thermal_energy_delivered
                + stsv.get_input_value(self.thermal_power_delivered_channel)
                * self.my_simulation_parameters.seconds_per_timestep
                * 1e-3
            )  # 1e-3 conversion J to kJ
        if self.thermal_power_chp_channel.source_output is not None:
            thermal_energy_delivered = (
                thermal_energy_delivered
                + stsv.get_input_value(self.thermal_power_chp_channel)
                * self.my_simulation_parameters.seconds_per_timestep
                * 1e-3
            )  # 1e-3 conversion J to kJ
        heatconsumption: float = self.calculate_heat_consumption(
            stsv=stsv,
            thermal_energy_delivered=thermal_energy_delivered,
        )
        stsv.set_output_value(self.power_from_water_storage_channel, heatconsumption)

        # constant heat loss of heat storage with the assumption that environment has 20°C = 293 K -> based on energy balance in kJ
        # heat gain due to heating of storage -> based on energy balance in kJ
        energy = self.state.energy_from_temperature()
        new_energy = (
            energy
            - (self.state.temperature_in_kelvin - 293)
            * self.surface
            * self.u_value
            * self.my_simulation_parameters.seconds_per_timestep
            * 1e-3
            - heatconsumption
            + thermal_energy_delivered
        )

        # convert new energy to new temperature
        self.state.set_temperature_from_energy(new_energy)

        # save outputs
        stsv.set_output_value(self.temperature_mean_channel, self.state.temperature_in_kelvin - 273.15)

    def calculate_heat_consumption(
        self,
        stsv: cp.SingleTimeStepValues,
        thermal_energy_delivered: float,
    ) -> float:
        """Calculates the heat consumption."""
        if self.use == lt.ComponentType.BOILER:
            # heat loss due to hot water consumption -> base on energy balance in kJ
            # 0.977 density of water in kg/l
            # 4.182 specific heat of water in kJ K^(-1) kg^(-1)
            return (
                stsv.get_input_value(self.water_consumption_channel)
                * (self.warm_water_temperature - self.drain_water_temperature)
                * 0.977
                * 4.182
            )
        if self.use == lt.ComponentType.BUFFER:
            heatconsumption = (
                stsv.get_input_value(self.heat_controller_target_percentage_channel)
                * self.power
                * self.my_simulation_parameters.seconds_per_timestep
                * 1e-3
            )  # 1e-3 conversion J to kJ
            available_energy = self.state.return_available_energy() + thermal_energy_delivered
            if heatconsumption > available_energy:
                heatconsumption = max(available_energy, 0)
            return heatconsumption
        raise Exception("Modular storage must be defined either as buffer or as boiler.")

    @staticmethod
    def get_cost_capex(config: StorageConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:
        """Returns investment cost, CO2 emissions and lifetime."""
        seconds_per_year = 365 * 24 * 60 * 60
        capex_per_simulated_period = (config.cost / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )
        device_co2_footprint_per_simulated_period = (config.co2_footprint / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )

        capex_cost_data_class = CapexCostDataClass(
            capex_investment_cost_in_euro=config.cost,
            device_co2_footprint_in_kg=config.co2_footprint,
            lifetime_in_years=config.lifetime,
            capex_investment_cost_for_simulated_period_in_euro=capex_per_simulated_period,
            device_co2_footprint_for_simulated_period_in_kg=device_co2_footprint_per_simulated_period,
            kpi_tag=KpiTagEnumClass.STORAGE_DOMESTIC_HOT_WATER
        )
        return capex_cost_data_class

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        # pylint: disable=unused-argument
        """Calculate OPEX costs, consisting of maintenance costs for DHW Storage."""
        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=self.calc_maintenance_cost(),
            co2_footprint_in_kg=0,
            consumption_in_kwh=0,
            loadtype=lt.LoadTypes.ANY,
            kpi_tag=KpiTagEnumClass.STORAGE_DOMESTIC_HOT_WATER
        )

        return opex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []



""" For setting the configuration of the household. """

# clean
from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import dataclass_json
from utspclient.helpers.lpgdata import ChargingStationSets
from utspclient.helpers.lpgpythonbindings import JsonReference


@dataclass_json
@dataclass
class SystemConfig:

    """Defines the configuration and sizing of all components considered in a modular household."""

    #: decision on the consideration of smart control for EV charging and heat pump
    surplus_control_considered: bool = False
    #: decision on the consideration of Photovoltaic Panel
    pv_included: bool = False
    #: peak power of the considered Photovoltaic Panel in Wp
    pv_peak_power: Optional[float] = 8e3
    #: decision on the consideration of Smart Control of Washing Machines, Dish Washers and Dryers
    smart_devices_included: bool = False
    #: decision on the consideration of a buffer storage for heating
    buffer_included: bool = False
    #: volume of the considered buffer storage in multiples of the default size
    buffer_volume: Optional[float] = 1.0  # in multiples of default
    #: decision on the consideration of battery
    battery_included: bool = False
    #: capacity of the considered battery in kWh
    battery_capacity: Optional[float] = 10.0  # in kWh
    #: decision on the consideration of heat pump
    heatpump_included: bool = False
    #: maximal power of the considered heat pump in multiples of the default
    heatpump_power: Optional[float] = 1.0  # in multiples of default
    #: decision on the consideration of combined heat and power - in this case a fuel cell
    chp_included: bool = False
    #: maximal power of the considered CHP in multiples of the default
    chp_power: Optional[float] = 0.5
    #: decision on the consideration of fuel cell + hydrogen storage + electrolyzer
    hydrogen_setup_included: bool = False
    #: maximal power of the considered fuel cell in kW (heat and electricity combined)
    fuel_cell_power: Optional[float] = 0.5
    #: size of the hydrogen storage in kg hydrogen
    h2_storage_size: Optional[float] = 100
    #: maximal power of the electroylzer in Watt
    electrolyzer_power: Optional[float] = 0.5
    #: decision on the consideration of an electriv vehicle
    ev_included: bool = False
    #: choice of charging station related to the options available in LoadProfileGenerator
    charging_station: JsonReference = field(
        default_factory=lambda: ChargingStationSets.Charging_At_Home_with_03_7_kW  # type: ignore
    )
